@@ -771,7 +771,8 @@ gok_spy_get_table_nodes (Accessible *search_root)
 
 	g_assert (Accessible_isTable (search_root));
 	table = Accessible_getTable (search_root);
-	flags.value = 0xff;
+	flags.value = ~0;
+	flags.data.inside_html_container = FALSE;
 	col_count = AccessibleTable_getNColumns (table);
 	if (col_count > 1)
 	{
@@ -863,7 +864,8 @@ gok_spy_get_children (Accessible *search_root)
 
 	g_assert (search_root);
 	/* TODO: should this recurse down, and should we check for 'interesting'-ness first ? */
-	flags.value = 0xff;
+	flags.value = ~0;
+	flags.data.inside_html_container = FALSE;
 	child_count = Accessible_getChildCount (search_root);
 	for (i = 0; i < child_count; ++i)
 	{
@@ -892,6 +894,7 @@ gok_spy_get_actionable_descendants (Accessible *accessible, GSList *nodes)
     GSList *ret_nodes;
     gboolean is_selection = Accessible_isSelection (accessible);
     flags.value = 0;
+    flags.data.inside_html_container = FALSE;
     flags.data.is_ui = TRUE;
 
     if (Accessible_isAction (accessible))
@@ -2065,6 +2068,25 @@ void gok_spy_process_focus (Accessible* accessible)
 	}
 	if (!has_menu) gok_spy_set_context_menu_accessible (NULL, 0);
 
+	/* 
+	 * if current active window is NULL, and this is a menu item, 
+	 * then we're in a popup menu of some kind; branch to the menus kbd
+	 */
+
+	if (accessible && (m_ptheWindowAccessible == NULL) && 
+	    gok_spy_is_menu_role (Accessible_getRole (accessible)))
+	{
+	    AccessibleNode *node = g_new0 (AccessibleNode, 1);
+	    gchar *name = Accessible_getName (accessible);
+	    node->link = -1;
+	    node->flags.value = 0;
+	    node->flags.data.is_menu = TRUE;
+	    node->pname = g_strdup (name ? name : "");
+	    gok_spy_accessible_ref (accessible);
+	    node->paccessible = accessible;
+	    gok_keyboard_branch_gui (node, GOK_SPY_SEARCH_CHILDREN);
+	}
+
 	gok_log_leave();
 } 
 
@@ -2246,28 +2268,31 @@ gok_spy_get_active_frame( )
 	return active_frame;
 }
 
-static Accessible*
-gok_spy_get_toolbar_item (Accessible *toolbar_child)
+/* recursive and exhaustive */
+static gboolean
+gok_spy_find_and_append_toolbar_items (Accessible *root, AccessibleNodeFlags flags)
 {
-	/* 
-	 * ROLE_TOOLBAR containers can have ROLE_PANEL or similarly useless children;
-	 * we need to drill down to the useful thingy i.e. pushbutton, etc.
-	 */
-	Accessible *child = toolbar_child, *tmp;
-	Accessible_ref (child);
-	while (child) 
+	Accessible *child, *tmp;
+	gint i, nchildren;
+	gboolean retval = FALSE;
+
+	if (gok_spy_is_menu_role (Accessible_getRole (root)))
 	{
-		if (Accessible_isAction (child)) 
-		{
-			return child;
-		}
-		else {
-			tmp = child;
-			child = Accessible_getChildAtIndex (child, 0);
-			Accessible_unref (tmp);
-		}
+	        return FALSE; /* make sure we continue searching independent of toolbars here */
 	}
-	return NULL;
+	else if (Accessible_isAction (root)) {
+		_priv_ui_nodes = gok_spy_append_node (_priv_ui_nodes, root, flags);
+		retval = TRUE;
+		/* uncomment if we don't care about children of actionables: */
+		/* return; */
+	}
+	nchildren = Accessible_getChildCount (root);
+	for (i = 0; i < nchildren; i++) {
+		child = Accessible_getChildAtIndex (root, i);
+		retval |= gok_spy_find_and_append_toolbar_items (child, flags);
+		Accessible_unref (child); 
+	}
+	return retval;
 }
 
 static GokSpyUIFlags 
@@ -2313,7 +2338,6 @@ gok_spy_search_component_list (Accessible *rootAccessible, GokSpyUIFlags keyboar
 		if (child)
 		{
 			gok_spy_accessible_implicit_ref (child);
-			flags.value = 0;
 			role = Accessible_getRole (child);
 			if (keyboard_ui_flags.data.menus && gok_spy_is_menu_role (role))
 			{
@@ -2328,16 +2352,10 @@ gok_spy_search_component_list (Accessible *rootAccessible, GokSpyUIFlags keyboar
 			else if (keyboard_ui_flags.data.toolbars && 
 				 parent_role == SPI_ROLE_TOOL_BAR)
 			{
-				Accessible *toolbar_item = gok_spy_get_toolbar_item (child);
 				if (keyboard_ui_flags.data.gui) 
 					flags.data.is_ui = TRUE; /* toolbar items are ui elements too */
 				ui_flags.data.toolbars = flags.data.is_toolbar_item = TRUE;
-				if (toolbar_item) 
-				{
-					_priv_ui_nodes = gok_spy_append_node (_priv_ui_nodes, toolbar_item, flags);
-					Accessible_unref (toolbar_item);
-				}
-				else
+				if (!gok_spy_find_and_append_toolbar_items (child, flags)) 
 				{
 				    ui_flags.value |= gok_spy_search_component_list (child, keyboard_ui_flags, context_flags).value;
 				}
@@ -2356,7 +2374,10 @@ gok_spy_search_component_list (Accessible *rootAccessible, GokSpyUIFlags keyboar
 			}
 			if (!flags.data.is_toolbar_item) /* toolbar items already added */
 			{
-				if (flags.value) 
+			        AccessibleNodeFlags non_context_flags = flags;
+				non_context_flags.data.inside_html_container = FALSE;
+
+				if (non_context_flags.value) 
 				{
 				    _priv_ui_nodes = gok_spy_append_node (_priv_ui_nodes, child, flags);
 				    /* special case for notebook tabs, whose children should also appear */
@@ -2407,7 +2428,8 @@ gok_spy_update_component_list (Accessible *rootAccessible, GokSpyUIFlags keyboar
 
 	gok_spy_free_nodes (_priv_ui_nodes);
 	_priv_ui_nodes = NULL;
-	
+
+	context_flags.value = 0;
 	ui_flags = gok_spy_search_component_list (rootAccessible, keyboard_ui_flags, context_flags);
 	/* _priv_ui_nodes is now re-assigned, by above call. */
 	gok_spy_resolve_namesakes (_priv_ui_nodes);
